"""Base class with Node specific helpers."""

from typing import cast

from ..base import ZWaveBase


class OZWNodeChildBase(ZWaveBase):
    """Base class for objects that are descendants of a Node object."""

    @property
    def node(self):
        """Return the node that this child belongs to."""
        from .node import OZWNode

        parent = self.parent

        while parent is not None and not isinstance(parent, OZWNode):
            parent = parent.parent

        if isinstance(parent, OZWNode):
            return cast(OZWNode, parent)

        raise RuntimeError("Object is not a descendant of a Node")

    def __repr__(self):
        """Return a representation of this object."""
        iden = f" {self.id}" if self.id else ""

        try:
            node = self.node.id
        except RuntimeError:
            node = "<missing> (bad!)"

        return f"<{type(self).__name__}{iden} (node: {node})>"



"""Base for all models."""
from abc import ABC
from typing import Any, Callable, Deque, Dict, Iterator, Optional, Type, Union

from .const import EMPTY_PAYLOAD, EVENT_PLACEHOLDER, LOGGER
from .options import OZWOptions


class ItemCollection:
    """Initialize an item collection."""

    def __init__(self, item_class: Type["ZWaveBase"]):
        """Initialize item collection instance."""
        self.parent: Optional["ZWaveBase"] = None
        self.topic_part: Optional[str] = None
        self.item_class = item_class
        self.collection: Dict[int, "ZWaveBase"] = {}

        assert item_class.EVENT_ADDED != EVENT_PLACEHOLDER
        assert item_class.EVENT_REMOVED != EVENT_PLACEHOLDER

    def adopt(self, parent: "ZWaveBase", topic_part: Optional[str]) -> None:
        """Adopt the item collection."""
        assert self.parent is None
        self.parent = parent
        self.topic_part = topic_part

    def get(self, item_id: int) -> Optional["ZWaveBase"]:
        """Return item in collection."""
        return self.collection.get(item_id)

    def process_message(self, topic: Deque[str], message: dict) -> None:
        """Process a new message."""
        item_id = int(topic.popleft())
        item = self.collection.get(item_id)
        added = False

        if item is None and message is EMPTY_PAYLOAD:
            return

        if item is None:
            topic_part = str(item_id)

            if self.topic_part is not None:
                topic_part = f"{self.topic_part}/{topic_part}"

            assert self.parent is not None
            item = self.collection[item_id] = self.item_class(
                self.parent.options, self.parent, topic_part, item_id
            )
            added = True

        if len(topic) == 0 and message is EMPTY_PAYLOAD:
            self.remove_and_notify(item_id)
            return

        item.process_message(topic, message)

        # Only notify after we process the message.
        if added:
            assert self.parent is not None
            self.parent.options.notify(self.item_class.EVENT_ADDED, item)

    def remove_and_notify(self, item_id: int) -> None:
        """Remove item from collection and fire remove events for all child objects."""
        item = self.collection[item_id]

        for collection in item.collections.values():
            if not isinstance(collection, ItemCollection):
                continue

            for item in list(collection):
                assert item.id is not None
                collection.remove_and_notify(item.id)

        assert self.parent is not None
        self.parent.options.notify(
            self.item_class.EVENT_REMOVED, self.collection.pop(item_id)
        )

    def __iter__(self) -> Iterator:
        """Return iterator over all items in this collection."""
        return iter(self.collection.values())


class ZWaveBase(ABC):
    """A base class for all models."""

    # Name the direct collection that lives underneath this object
    # but is not named in the topic. A message to /openzwave/1 will
    # be interpreted as if sent to /openzwave/<DIRECT_COLLECTION>/1
    DIRECT_COLLECTION: Optional[str] = None

    # Default value of this object. If untouched, all messages for child objects
    # will be held until information for this object has been received.
    DEFAULT_VALUE: Optional[dict] = EMPTY_PAYLOAD

    # Use in case there is a special plural name of this class.
    PLURAL_NAME: Optional[str] = None

    EVENT_ADDED = EVENT_PLACEHOLDER
    EVENT_CHANGED = EVENT_PLACEHOLDER
    EVENT_REMOVED = EVENT_PLACEHOLDER

    def __init__(
        self,
        options: OZWOptions,
        parent: Optional["ZWaveBase"],
        topic_part: str,
        item_id: Optional[int],
    ):
        """Initialize the model."""
        # Runtime options
        self.options = options

        # Parent object
        self.parent = parent

        # Part of the topic that instantiated this object.
        self.topic_part = topic_part

        # Identifier of this object
        self.id = item_id

        # Models that live under this model
        self.collections: Dict[
            str, Union[ItemCollection, "ZWaveBase", "DiscardMessages", "EventMessages"]
        ] = {}

        # The data this object holds
        self.data = self.DEFAULT_VALUE

        # Messages for children that are held until data is received
        self.pending_messages: Optional[list] = None

        assert self.EVENT_CHANGED != EVENT_PLACEHOLDER

        # Process collections
        for item_type, collection in self.create_collections().items():
            if isinstance(collection, type):  # OZWBase
                self.collections[item_type] = collection(
                    self.options, self, item_type, None
                )
                setattr(
                    self, f"get_{item_type}", create_getter(self.collections[item_type])
                )
                continue

            if not isinstance(collection, ItemCollection):
                self.collections[item_type] = collection
                continue

            setattr(self, f"get_{item_type}", collection.get)

            plural_name = collection.item_class.PLURAL_NAME or f"{item_type}s"
            if hasattr(self, plural_name):
                raise RuntimeError(
                    f"Cannot add {plural_name} function to {type(self).__name__}. Already exists."
                )
            setattr(self, plural_name, collection.collection.values)

            if item_type == self.DIRECT_COLLECTION:
                coll_topic_part: Optional[str] = None
            else:
                coll_topic_part = item_type

            collection.adopt(self, coll_topic_part)
            self.collections[item_type] = collection

    @property
    def topic(self) -> str:
        """Return topic of this object."""
        if self.parent is None:
            # Cut off the trailing slash
            return self.options.topic_prefix[:-1]

        return f"{self.parent.topic}/{self.topic_part}"

    def create_collections(
        self,
    ) -> Dict[
        str,
        Union[ItemCollection, Type["ZWaveBase"], "DiscardMessages", "EventMessages"],
    ]:
        """Create collections that this type supports.

        Each collection is either ItemCollection or a class derived from ZWaveBase.
        """
        # pylint: disable=no-self-use
        return {}

    def process_message(self, topic: Deque[str], message: dict) -> None:
        """Process a new message."""
        if len(topic) == 0:
            is_init_msg = self.data is EMPTY_PAYLOAD
            self.data = message

            if not is_init_msg:
                self.options.notify(self.EVENT_CHANGED, self)
                return

            # Process all messages for the children.
            if self.pending_messages is not None:
                for pend_topic, pend_message in self.pending_messages:
                    self.process_message(pend_topic, pend_message)
                self.pending_messages = None

            return

        # If this object has not been initialized, queue up messages.
        if self.data is EMPTY_PAYLOAD:
            if self.pending_messages is None:
                self.pending_messages = []
            self.pending_messages.append((topic, message))
            return

        if topic[0] in self.collections:
            collection_type = topic.popleft()

        elif self.DIRECT_COLLECTION and topic[0].isnumeric():
            collection_type = self.DIRECT_COLLECTION

        else:
            self._warn_cannot_handle(topic, message)
            return

        self.collections[collection_type].process_message(topic, message)

    def _warn_cannot_handle(self, topic: Deque[str], message: dict) -> None:
        LOGGER.warning(
            "%s cannot process message %s: %s",
            type(self).__name__,
            f"{self.topic}/{'/'.join(topic)}",
            message,
        )

    def __repr__(self) -> str:
        """Return a representation of this object."""
        iden = f" {self.id}" if self.id else ""
        return f"<{type(self).__name__}{iden}>"


class DiscardMessages:
    """Class that discards all messages sent to it."""

    def process_message(self, topic: Deque[str], message: dict) -> None:
        """Process incoming message."""


class EventMessages:
    """Class that converts messages to events."""

    def __init__(
        self,
        options: OZWOptions,
        event: str,
        type_extractor: Callable[[Deque[str], dict], str],
    ):
        """Initialize EventMessages."""
        self.options = options
        self.event = event
        self.type_extractor = type_extractor

    def process_message(self, topic: Deque[str], message: dict) -> None:
        """Process incoming message."""
        event_type = self.type_extractor(topic, message)
        self.options.notify(self.event, {"event": event_type, "data": message})


def create_getter(obj: Any) -> Callable:
    """Return a function that returns an object.

    Workaround for not being able to create lambdas that refer to variables in the
    current iteration of the loop.
    """
    return lambda: obj



"""Root Manager object."""
import json
from collections import deque
from typing import TYPE_CHECKING, Dict, Optional, Type, Union

from .base import ItemCollection, ZWaveBase
from .const import EMPTY_PAYLOAD
from .models.instance import OZWInstance
from .options import OZWOptions

if TYPE_CHECKING:
    from .base import DiscardMessages, EventMessages  # noqa: F401


class OZWManager(ZWaveBase):
    """Manager that holds the OZW instances connected to MQTT."""

    DIRECT_COLLECTION = "instance"
    DEFAULT_VALUE: Optional[dict] = None
    EVENT_CHANGED = "manager_placeholder_event"

    def __init__(self, options: OZWOptions):
        """Initialize class."""
        super().__init__(options, None, options.topic_prefix, None)

    def create_collections(
        self,
    ) -> Dict[
        str,
        Union[ItemCollection, Type["ZWaveBase"], "DiscardMessages", "EventMessages"],
    ]:
        """Create collections that the manager supports."""
        return {"instance": ItemCollection(OZWInstance)}

    def receive_message(self, topic: str, message: str) -> None:
        """Receive an MQTT message."""
        assert topic.startswith(self.options.topic_prefix)

        topic_parts_raw = topic[len(self.options.topic_prefix) :].split("/")
        if topic_parts_raw[-1] == "":
            topic_parts_raw.pop()
        topic_parts = deque(topic_parts_raw)

        if message == "":
            payload = EMPTY_PAYLOAD
        else:
            payload = json.loads(message)

        self.process_message(topic_parts, payload)



"""Model for the Node instance level."""
from typing import Optional

from ..base import ItemCollection
from ..const import (
    EVENT_NODE_INSTANCE_ADDED,
    EVENT_NODE_INSTANCE_CHANGED,
    EVENT_NODE_INSTANCE_REMOVED,
    CommandClass,
    ValueIndex,
)
from .command_class import OZWCommandClass
from .node_child_base import OZWNodeChildBase
from .value import OZWValue


class OZWNodeInstance(OZWNodeChildBase):
    """Model for Node Instance."""

    EVENT_ADDED = EVENT_NODE_INSTANCE_ADDED
    EVENT_CHANGED = EVENT_NODE_INSTANCE_CHANGED
    EVENT_REMOVED = EVENT_NODE_INSTANCE_REMOVED

    @property
    def instance(self) -> int:
        """Return Instance."""
        return self.data.get("Instance")

    @property
    def time_stamp(self) -> int:
        """Return TimeStamp."""
        return self.data.get("TimeStamp")

    def create_collections(self):
        """Create collections that Node supports."""
        return {"commandclass": ItemCollection(OZWCommandClass)}

    def get_command_class(
        self, command_class_id: CommandClass
    ) -> Optional[OZWCommandClass]:
        """Return a specific CommandClass on this NodeInstance (if exists)."""
        # pylint: disable=no-member
        for command_class in self.commandclasses():
            if command_class.command_class_id == command_class_id:
                return command_class
        return None

    def has_command_class(self, command_class_id: CommandClass) -> bool:
        """Determine if the node has the given CommandClass."""
        return self.get_command_class(command_class_id) is not None

    def get_value(
        self, command_class_id: CommandClass, value_index: ValueIndex
    ) -> Optional[OZWValue]:
        """Return a specific OZWValue on this node (if exists)."""
        command_class = self.get_command_class(command_class_id)
        return command_class.get_value(value_index) if command_class else None

    def has_value(
        self, command_class_id: CommandClass, value_index: ValueIndex
    ) -> bool:
        """Determine if this NodeInstance has the given OZWValue."""
        return self.get_value(command_class_id, value_index) is not None



"""Model for the Value."""
from typing import cast

from ..const import (
    EVENT_VALUE_ADDED,
    EVENT_VALUE_CHANGED,
    EVENT_VALUE_REMOVED,
    CommandClass,
    ValueGenre,
    ValueType,
    ValueIndex,
)
from .node_child_base import OZWNodeChildBase


class OZWValue(OZWNodeChildBase):
    """Representation of an OpenZWave Value object."""

    EVENT_ADDED = EVENT_VALUE_ADDED
    EVENT_CHANGED = EVENT_VALUE_CHANGED
    EVENT_REMOVED = EVENT_VALUE_REMOVED

    @property
    def label(self) -> str:
        """Return Label."""
        return self.data.get("Label")

    @property
    def value(self) -> int:
        """Return Value."""
        return self.data.get("Value")

    @property
    def units(self) -> str:
        """Return Units."""
        return self.data.get("Units")

    @property
    def min(self) -> int:
        """Return Min."""
        return self.data.get("Min")

    @property
    def max(self) -> int:
        """Return Max."""
        return self.data.get("Max")

    @property
    def type(self) -> ValueType:
        """Return Type."""
        try:
            return ValueType(self.data.get("Type"))
        except ValueError:
            return ValueType.UNKNOWN

    @property
    def instance(self) -> int:
        """Return Instance."""
        return self.data.get("Instance")

    @property
    def command_class(self) -> CommandClass:
        """Return CommandClass this value belongs to."""
        return self.parent.command_class_id

    @property
    def index(self) -> ValueIndex:
        """Return Index."""
        try:
            # TODO: we can make this prettier by returning the
            # ValueIndex belonging to this specific CC
            # Now it will return the first value matching this int value
            # Nothing breaks but it isn't very pretty
            return ValueIndex(self.data.get("Index"))
        except ValueError:
            return ValueIndex.UNKNOWN

    @property
    def genre(self) -> ValueGenre:
        """Return Genre."""
        try:
            return ValueGenre(self.data.get("Genre"))
        except ValueError:
            return ValueGenre.UNKNOWN

    @property
    def help(self) -> str:
        """Return Help."""
        return self.data.get("Help")

    @property
    def value_id_key(self) -> int:
        """Return ValueIDKey."""
        return self.data.get("ValueIDKey")

    @property
    def read_only(self) -> bool:
        """Return ReadOnly."""
        return self.data.get("ReadOnly")

    @property
    def write_only(self) -> bool:
        """Return WriteOnly."""
        return self.data.get("WriteOnly")

    @property
    def value_set(self) -> bool:
        """Return ValueSet."""
        return self.data.get("ValueSet")

    @property
    def value_polled(self) -> bool:
        """Return ValuePolled."""
        return self.data.get("ValuePolled")

    @property
    def change_verified(self) -> bool:
        """Return ChangeVerified."""
        return self.data.get("ChangeVerified")

    @property
    def event(self) -> str:
        """Return Event."""
        return self.data.get("Event")

    @property
    def time_stamp(self) -> int:
        """Return TimeStamp."""
        return self.data.get("TimeStamp")

    @property
    def ozw_instance(self):
        """Return OZWInstance this value belongs to."""
        from .instance import OZWInstance

        parent = self.parent
        while parent is not None and not isinstance(parent, OZWInstance):
            parent = parent.parent

        if isinstance(parent, OZWInstance):
            return cast(OZWInstance, parent)

    def send_value(self, new_value):
        """Send an updated value to MQTT."""
        instance_id = self.ozw_instance.id
        full_topic = f"{self.options.topic_prefix}{instance_id}/command/setvalue/"
        payload = {"ValueIDKey": self.value_id_key, "Value": new_value}
        self.options.send_message(full_topic, payload)



"""Model for OZW Instance Status."""
from .. import base
from ..const import EVENT_INSTANCE_STATUS_CHANGED


class OZWInstanceStatus(base.ZWaveBase):
    """Model for OZW Instance Status."""

    EVENT_CHANGED = EVENT_INSTANCE_STATUS_CHANGED

    @property
    def status(self):
        """Return current status of this OZW Instance."""
        return self.data.get("Status")

    @property
    def home_id(self):
        """Return the homeID of this OZW Instance."""
        return self.data.get("homeID")

    @property
    def manufacturer_specific_db_ready(self) -> bool:
        """Return ManufacturerSpecificDBReady."""
        return self.data.get("ManufacturerSpecificDBReady")

    @property
    def time_stamp(self) -> int:
        """Return TimeStamp."""
        return self.data.get("TimeStamp")

    @property
    def openzwave_version(self) -> str:
        """Return OpenZWave_Version."""
        return self.data.get("OpenZWave_Version")

    @property
    def ozw_deamon_version(self) -> str:
        """Return OZWDeamon_Version."""
        return self.data.get("OZWDeamon_Version")

    @property
    def qt_openzwave_version(self) -> str:
        """Return QTOpenZWave_Version."""
        return self.data.get("QTOpenZWave_Version")

    @property
    def qt_version(self) -> str:
        """Return QT_Version."""
        return self.data.get("QT_Version")

    @property
    def get_controller_node_id(self) -> int:
        """Return getControllerNodeId."""
        return self.data.get("getControllerNodeId")

    @property
    def get_suc_node_id(self) -> int:
        """Return getSUCNodeId."""
        return self.data.get("getSUCNodeId")

    @property
    def is_primary_controller(self) -> bool:
        """Return isPrimaryController."""
        return self.data.get("isPrimaryController")

    @property
    def is_bridge_controller(self) -> bool:
        """Return isBridgeController."""
        return self.data.get("isBridgeController")

    @property
    def has_extended_tx_statistics(self) -> bool:
        """Return hasExtendedTXStatistics."""
        return self.data.get("hasExtendedTXStatistics")

    @property
    def get_controller_library_version(self) -> str:
        """Return getControllerLibraryVersion."""
        return self.data.get("getControllerLibraryVersion")

    @property
    def get_controller_library_type(self) -> str:
        """Return getControllerLibraryType."""
        return self.data.get("getControllerLibraryType")

    @property
    def get_controller_path(self) -> str:
        """Return getControllerPath."""
        return self.data.get("getControllerPath")



"""Model for the Node Assocations level."""
from ..const import (
    EVENT_NODE_ASSOCIATION_ADDED,
    EVENT_NODE_ASSOCIATION_CHANGED,
    EVENT_NODE_ASSOCIATION_REMOVED,
)
from .node_child_base import OZWNodeChildBase


class OZWNodeAssocation(OZWNodeChildBase):
    """Model for Node Assocations."""

    EVENT_ADDED = EVENT_NODE_ASSOCIATION_ADDED
    EVENT_CHANGED = EVENT_NODE_ASSOCIATION_CHANGED
    EVENT_REMOVED = EVENT_NODE_ASSOCIATION_REMOVED

    @property
    def name(self) -> str:
        """Return Name."""
        return self.data.get("Name")

    @property
    def help(self) -> str:
        """Return Help."""
        return self.data.get("Help")

    @property
    def max_associations(self) -> int:
        """Return MaxAssociations."""
        return self.data.get("MaxAssociations")

    @property
    def members(self) -> list:
        """Return Members."""
        return self.data.get("Members")

    @property
    def time_stamp(self) -> int:
        """Return TimeStamp."""
        return self.data.get("TimeStamp")



"""Model for the Instance statistics."""
from ..const import EVENT_INSTANCE_STATISTICS_CHANGED
from .node_child_base import OZWNodeChildBase


class OZWInstanceStatistics(OZWNodeChildBase):
    """Model for OZW Instance statistics."""

    EVENT_CHANGED = EVENT_INSTANCE_STATISTICS_CHANGED

    @property
    def sof_cnt(self) -> int:
        """Return SOFCnt."""
        return self.data.get("SOFCnt")

    @property
    def ack_waiting(self) -> int:
        """Return ACKWaiting."""
        return self.data.get("ACKWaiting")

    @property
    def read_aborts(self) -> int:
        """Return readAborts."""
        return self.data.get("readAborts")

    @property
    def bad_checksum(self) -> int:
        """Return badChecksum."""
        return self.data.get("badChecksum")

    @property
    def read_cnt(self) -> int:
        """Return readCnt."""
        return self.data.get("readCnt")

    @property
    def write_cnt(self) -> int:
        """Return writeCnt."""
        return self.data.get("writeCnt")

    @property
    def can_cnt(self) -> int:
        """Return CANCnt."""
        return self.data.get("CANCnt")

    @property
    def nak_cnt(self) -> int:
        """Return NAKCnt."""
        return self.data.get("NAKCnt")

    @property
    def ack_cnt(self) -> int:
        """Return ACKCnt."""
        return self.data.get("ACKCnt")

    @property
    def oof_cnt(self) -> int:
        """Return OOFCnt."""
        return self.data.get("OOFCnt")

    @property
    def dropped(self) -> int:
        """Return dropped."""
        return self.data.get("dropped")

    @property
    def retries(self) -> int:
        """Return retries."""
        return self.data.get("retries")

    @property
    def callbacks(self) -> int:
        """Return callbacks."""
        return self.data.get("callbacks")

    @property
    def badroutes(self) -> int:
        """Return badroutes."""
        return self.data.get("badroutes")

    @property
    def noack(self) -> int:
        """Return noack."""
        return self.data.get("noack")

    @property
    def netbusy(self) -> int:
        """Return netbusy."""
        return self.data.get("netbusy")

    @property
    def notidle(self) -> int:
        """Return notidle."""
        return self.data.get("notidle")

    @property
    def txverified(self) -> int:
        """Return txverified."""
        return self.data.get("txverified")

    @property
    def nondelivery(self) -> int:
        """Return nondelivery."""
        return self.data.get("nondelivery")

    @property
    def routedbusy(self) -> int:
        """Return routedbusy."""
        return self.data.get("routedbusy")

    @property
    def broadcast_read_cnt(self) -> int:
        """Return broadcastReadCnt."""
        return self.data.get("broadcastReadCnt")

    @property
    def broadcast_write_cnt(self) -> int:
        """Return broadcastWriteCnt."""
        return self.data.get("broadcastWriteCnt")



"""Model for a OZW Node."""
from typing import Iterable, List, Optional

from ..base import ItemCollection, ZWaveBase
from ..const import (
    EVENT_NODE_ADDED,
    EVENT_NODE_CHANGED,
    EVENT_NODE_REMOVED,
    CommandClass,
    ValueIndex,
)
from .command_class import OZWCommandClass
from .node_association import OZWNodeAssocation
from .node_instance import OZWNodeInstance
from .node_statistics import OZWNodeStatistics
from .value import OZWValue


class OZWNode(ZWaveBase):
    """Model for a Z-Wave Node."""

    EVENT_ADDED = EVENT_NODE_ADDED
    EVENT_CHANGED = EVENT_NODE_CHANGED
    EVENT_REMOVED = EVENT_NODE_REMOVED

    @property
    def node_id(self) -> int:
        """Return NodeID."""
        return self.data.get("NodeID")

    @property
    def node_query_stage(self) -> str:
        """Return NodeQueryStage."""
        return self.data.get("NodeQueryStage")

    @property
    def is_listening(self) -> bool:
        """Return isListening."""
        return self.data.get("isListening")

    @property
    def is_flirs(self) -> bool:
        """Return isFlirs."""
        return self.data.get("isFlirs")

    @property
    def is_beaming(self) -> bool:
        """Return isBeaming."""
        return self.data.get("isBeaming")

    @property
    def is_routing(self) -> bool:
        """Return isRouting."""
        return self.data.get("isRouting")

    @property
    def is_securityv1(self) -> bool:
        """Return isSecurityv1."""
        return self.data.get("isSecurityv1")

    @property
    def is_zwave_plus(self) -> bool:
        """Return isZWavePlus."""
        return self.data.get("isZWavePlus")

    @property
    def is_nif_recieved(self) -> bool:
        """Return isNIFRecieved."""
        return self.data.get("isNIFRecieved")

    @property
    def is_awake(self) -> bool:
        """Return isAwake."""
        return self.data.get("isAwake")

    @property
    def is_failed(self) -> bool:
        """Return isFailed."""
        return self.data.get("isFailed")

    @property
    def meta_data(self) -> dict:
        """Return MetaData."""
        return self.data.get("MetaData")

    @property
    def event(self) -> str:
        """Return Event."""
        return self.data.get("Event")

    @property
    def time_stamp(self) -> int:
        """Return TimeStamp."""
        return self.data.get("TimeStamp")

    @property
    def node_manufacturer_name(self) -> str:
        """Return NodeManufacturerName."""
        return self.data.get("NodeManufacturerName")

    @property
    def node_product_name(self) -> str:
        """Return NodeProductName."""
        return self.data.get("NodeProductName")

    @property
    def node_basic_string(self) -> str:
        """Return NodeBasicString."""
        return self.data.get("NodeBasicString")

    @property
    def node_basic(self) -> int:
        """Return NodeBasic."""
        return self.data.get("NodeBasic")

    @property
    def node_generic_string(self) -> str:
        """Return NodeGenericString."""
        return self.data.get("NodeGenericString")

    @property
    def node_generic(self) -> int:
        """Return NodeGeneric."""
        return self.data.get("NodeGeneric")

    @property
    def node_specific_string(self) -> int:
        """Return NodeSpecificString."""
        return self.data.get("NodeSpecificString")

    @property
    def node_specific(self) -> int:
        """Return NodeSpecific."""
        return self.data.get("NodeSpecific")

    @property
    def node_manufacturer_id(self) -> str:
        """Return NodeManufacturerID."""
        return self.data.get("NodeManufacturerID")

    @property
    def node_product_type(self) -> str:
        """Return NodeProductType."""
        return self.data.get("NodeProductType")

    @property
    def node_product_id(self) -> str:
        """Return NodeProductID."""
        return self.data.get("NodeProductID")

    @property
    def node_baud_rate(self) -> int:
        """Return NodeBaudRate."""
        return self.data.get("NodeBaudRate")

    @property
    def node_version(self) -> int:
        """Return NodeVersion."""
        return self.data.get("NodeVersion")

    @property
    def node_groups(self) -> int:
        """Return NodeGroups."""
        return self.data.get("NodeGroups")

    @property
    def node_name(self) -> str:
        """Return NodeName."""
        return self.data.get("NodeName")

    @property
    def node_location(self) -> str:
        """Return NodeLocation."""
        return self.data.get("NodeLocation")

    @property
    def node_device_type_string(self) -> str:
        """Return NodeDeviceTypeString."""
        return self.data.get("NodeDeviceTypeString")

    @property
    def node_device_type(self) -> int:
        """Return NodeDeviceType."""
        return self.data.get("NodeDeviceType")

    @property
    def node_role(self) -> int:
        """Return NodeRole."""
        return self.data.get("NodeRole")

    @property
    def node_role_string(self) -> str:
        """Return NodeRoleString."""
        return self.data.get("NodeRoleString")

    @property
    def node_plus_type(self) -> int:
        """Return NodePlusType."""
        return self.data.get("NodePlusType")

    @property
    def node_plus_type_string(self) -> str:
        """Return NodePlusTypeString."""
        return self.data.get("NodePlusTypeString")

    @property
    def neighbors(self) -> List[int]:
        """Return Neighbors."""
        return self.data.get("Neighbors")

    def values(self) -> Iterable["OZWValue"]:
        """Iterate over all OZWValue child items."""
        # pylint: disable=no-member
        return (
            value
            for instance in self.instances()
            for cc in instance.commandclasses()
            for value in cc.values()
        )

    def create_collections(self):
        """Create collections that Node supports."""
        return {
            "instance": ItemCollection(OZWNodeInstance),
            "association": ItemCollection(OZWNodeAssocation),
            "statistics": OZWNodeStatistics,
        }

    def get_command_class(
        self, command_class_id: CommandClass, instance_id: Optional[int] = None
    ) -> Optional[OZWCommandClass]:
        """Return a specific CommandClass on this node (if exists)."""
        # pylint: disable=no-member
        for instance in self.instances():
            if instance_id is not None and instance.instance != instance_id:
                continue
            return instance.get_command_class(command_class_id)
        return None

    def has_command_class(
        self, command_class_id: CommandClass, instance_id: Optional[int] = None
    ) -> bool:
        """Determine if the node has the given CommandClass."""
        return self.get_command_class(command_class_id, instance_id) is not None

    def get_value(
        self,
        command_class_id: CommandClass,
        value_index: ValueIndex,
        instance_id: Optional[int] = None,
    ) -> Optional[OZWValue]:
        """Return a specific OZWValue on this node (if exists)."""
        command_class = self.get_command_class(command_class_id, instance_id)
        return command_class.get_value(value_index) if command_class else None

    def has_value(
        self, command_class_id: CommandClass, instance_id: Optional[int] = None
    ) -> bool:
        """Determine if the node has the given CommandClass."""
        return self.get_command_class(command_class_id, instance_id) is not None



"""Model for the OZW instance level."""
from typing import Dict, Optional, Type, Union

from .. import base
from ..const import (
    EVENT_INSTANCE_ADDED,
    EVENT_INSTANCE_CHANGED,
    EVENT_INSTANCE_EVENT,
    EVENT_INSTANCE_REMOVED,
)
from .instance_statistics import OZWInstanceStatistics
from .instance_status import OZWInstanceStatus
from .node import OZWNode


class OZWInstance(base.ZWaveBase):
    """Model for the OZW instance level."""

    DEFAULT_VALUE: Optional[dict] = None

    EVENT_ADDED = EVENT_INSTANCE_ADDED
    EVENT_CHANGED = EVENT_INSTANCE_CHANGED
    EVENT_REMOVED = EVENT_INSTANCE_REMOVED

    def create_collections(
        self,
    ) -> Dict[
        str,
        Union[
            base.ItemCollection,
            Type[base.ZWaveBase],
            base.DiscardMessages,
            base.EventMessages,
        ],
    ]:
        """Create collections that Node supports."""
        return {
            "node": base.ItemCollection(OZWNode),
            "status": OZWInstanceStatus,
            "statistics": OZWInstanceStatistics,
            "command": base.DiscardMessages(),
            "event": base.EventMessages(
                self.options, EVENT_INSTANCE_EVENT, lambda topic, data: topic[0]
            ),
        }

    def send_command(self, command: str, payload: Optional[dict] = None) -> None:
        """Send command to the OZW instance."""
        if payload is None:
            payload = {}
        topic_prefix = self.options.topic_prefix
        full_topic = f"{topic_prefix}{self.id}/command/{command}/"
        self.options.send_message(full_topic, payload)

    # Shortcut methods to some common used (global) controller commands
    # https://github.com/OpenZWave/qt-openzwave/blob/master/docs/MQTT.md#mqtt-commands

    def add_node(self, secure: bool = False) -> None:
        """Enter inclusion mode on the controller."""
        self.send_command("addnode", {"secure": secure})

    def remove_node(self) -> None:
        """Enter exclusion mode on the controller."""
        self.send_command("removenode")

    def refresh_node(self, node_id: int) -> None:
        """Force OZW to re-interview a device."""
        self.send_command("refreshnodeinfo", {"node": node_id})

    def remove_failed_node(self, node_id: int) -> None:
        """Remove a failed node from the controller."""
        self.send_command("removefailednode", {"node": node_id})

    def replace_failed_node(self, node_id: int) -> None:
        """Replace a failed node from the controller with a new device."""
        self.send_command("replacefailednode", {"node": node_id})

    def heal_node(self, node_id: int) -> None:
        """Ask a Node to recalculate its neighbors and routes to other devices."""
        self.send_command("healnetworknode", {"node": node_id})

    def cancel_controller_command(self) -> None:
        """Cancel in Controller Commands that are in progress."""
        self.send_command("cancelcontrollercommand")



"""Model for the CommandClass."""
from typing import Optional

from ..base import ItemCollection
from ..const import (
    EVENT_COMMAND_CLASS_ADDED,
    EVENT_COMMAND_CLASS_CHANGED,
    EVENT_COMMAND_CLASS_REMOVED,
    LOGGER,
    CommandClass,
    ValueIndex,
)
from .node_child_base import OZWNodeChildBase
from .value import OZWValue


class OZWCommandClass(OZWNodeChildBase):
    """Model for the OZW CommandClass."""

    EVENT_ADDED = EVENT_COMMAND_CLASS_ADDED
    EVENT_CHANGED = EVENT_COMMAND_CLASS_CHANGED
    EVENT_REMOVED = EVENT_COMMAND_CLASS_REMOVED

    PLURAL_NAME = "commandclasses"

    @property
    def instance(self) -> int:
        """Return Instance."""
        return self.data.get("Instance")

    @property
    def command_class_id(self) -> CommandClass:
        """Return CommandClassId as CommandClass Enum."""
        try:
            return CommandClass(self.data.get("CommandClassId"))
        except ValueError:
            LOGGER.warning(
                "Unknown CommandClass found: %s", self.data.get("CommandClassId")
            )
            return CommandClass.UNKNOWN

    @property
    def command_class(self) -> str:
        """Return string/label representation of this CommandClass."""
        # TODO: This needs changing to a better name once fixed upstream in the daemon
        return self.data.get("CommandClass")

    @property
    def time_stamp(self) -> int:
        """Return TimeStamp."""
        return self.data.get("TimeStamp")

    def create_collections(self):
        """Create collections that Node supports."""
        return {"value": ItemCollection(OZWValue)}

    def get_value(self, value_index: ValueIndex) -> Optional[OZWValue]:
        """Return a specific OZWValue on this CommandClass (if exists)."""
        # pylint: disable=no-member
        for value in self.values():
            if value.index == value_index:
                return value
        return None

    def has_value(self, value_index: ValueIndex) -> bool:
        """Determine if the CommandClass has the given ValueIndex."""
        return self.get_value(value_index) is not None



"""Options for the OZW MQTT Connection."""
from typing import TYPE_CHECKING, Callable, Dict, List, Union

if TYPE_CHECKING:
    from .base import ZWaveBase  # noqa: F401


class OZWOptions:
    """OZW Options class."""

    def __init__(
        self,
        send_message: Callable[[str, Union[str, dict]], None],
        topic_prefix: str = "OpenZWave/",
    ):
        """Initialize class."""
        self.send_message = send_message
        self.topic_prefix = topic_prefix
        self.listeners: Dict[str, List[Callable[[Union[dict, "ZWaveBase"]], None]]] = {}

        # Make sure topic prefix ends in a slash
        assert topic_prefix[-1] == "/"

    def listen(
        self, event: str, listener: Callable[[Union[dict, "ZWaveBase"]], None]
    ) -> None:
        """Attach listener for events."""
        self.listeners.setdefault(event, []).append(listener)

    def notify(self, event: str, data: Union[dict, "ZWaveBase"]) -> None:
        """Notify listeners of a new event."""
        for listener in self.listeners.get(event, []):
            listener(data)



"""Provide tests for the manager."""


def test_receive_message(mgr):
    """Test receive message processing."""

    def mock_process_message(topic_parts, payload):
        """Test data comes in ok."""
        assert list(topic_parts) == ["1", "node", "2", "value", "3"]
        assert payload == {"mock": "payload"}

    mgr.process_message = mock_process_message

    mgr.receive_message("OpenZWave/1/node/2/value/3/", '{"mock":"payload"}')



"""Provide tests for the model base."""
from collections import deque
from unittest.mock import patch

import pytest

from openzwavemqtt import base


class Level3(base.ZWaveBase):
    """Represent a level 3 descendant."""

    EVENT_ADDED = "level3_added"
    EVENT_CHANGED = "level3_change"
    EVENT_REMOVED = "level3_removed"

    @property
    def hello(self):
        """Return hello data."""
        return self.data.get("hello")


class Level2(base.ZWaveBase):
    """Represent a level 2 descendant."""

    DIRECT_COLLECTION = "level3"
    EVENT_ADDED = "level2_added"
    EVENT_CHANGED = "level2_change"
    EVENT_REMOVED = "level2_removed"

    PLURAL_NAME = "level_twos"

    def create_collections(self):
        """Create collections."""
        return {"level3": base.ItemCollection(Level3)}


class Level1(base.ZWaveBase):
    """Represent a level 1 descendant."""

    DIRECT_COLLECTION = "level2"
    EVENT_ADDED = "level1_added"
    EVENT_CHANGED = "level1_change"
    EVENT_REMOVED = "level1_removed"

    def create_collections(self):
        """Create collections."""
        return {"level2": base.ItemCollection(Level2)}


@pytest.fixture(name="level1")
def level1_fixture(options):
    """Return a level 1 descendant."""
    return Level1(options, None, None, None)


def test_direct_collection(level1, caplog):
    """Test direct collection."""
    level1.process_message(deque(), {"info": 1})
    level1.process_message(deque(["2"]), {"info": 1})
    level1.process_message(deque(["2", "3"]), {"hello": 1})
    assert level1.get_level2(2).get_level3(3).hello == 1

    # Only works on numbers
    level1.process_message(deque(["2", "a"]), {"hello": 1})
    assert level1.get_level2(2).get_level3("a") is None
    assert "cannot process message" in caplog.text


def test_pending_messages(level1, options):
    """Test pending messages."""
    events = []
    options.notify = lambda event, data: events.append(event)

    # Only message for level3 has been received, level2 is none
    level1.process_message(deque(["2", "3"]), {"hello": 1})
    assert level1.get_level2(2) is None
    assert events == []

    # Message for level2, level3 received, level1 still None
    level1.process_message(deque(["2"]), {"hello": 1})
    assert level1.get_level2(2) is None
    assert events == []

    # Level 1 receives data, process all child messages.
    level1.process_message(deque(), {"info": 1})
    assert level1.get_level2(2).get_level3(3).hello == 1
    assert events == ["level2_added", "level3_added"]


def test_recursive_remove(level1, options):
    """Test recursive remove."""
    events = []

    level1.process_message(deque(), {"info": 1})
    level1.process_message(deque(["2"]), {"info": 1})
    level1.process_message(deque(["2", "3"]), {"hello": 1})

    options.notify = lambda event, data: events.append(event)
    level1.process_message(deque(["2"]), base.EMPTY_PAYLOAD)

    assert events == ["level3_removed", "level2_removed"]


def test_topic(options):
    """Test topic property."""

    class Level4(base.ZWaveBase):
        """Represent a level 4 descendant."""

        EVENT_ADDED = "level4_added"
        EVENT_CHANGED = "level4_change"
        EVENT_REMOVED = "level4_removed"

    class Level3Statistics(base.ZWaveBase):
        """Represent level 3 statistics."""

        EVENT_CHANGED = "level3statistics_change"

    # Patch in a non-direct collection
    with patch.object(
        Level3,
        "create_collections",
        lambda _: {
            "level4": base.ItemCollection(Level4),
            "statistics": Level3Statistics,
        },
    ):
        level1 = Level1(options, None, None, None)
        level1.process_message(deque(), {"info": 1})
        level1.process_message(deque(["2"]), {"info": 1})
        level1.process_message(deque(["2", "3"]), {"hello": 1})
        level1.process_message(deque(["2", "3", "level4", "4"]), {"hello": 1})
        level1.process_message(deque(["2", "3", "statistics"]), {"hello": 1})

    assert (
        level1.get_level2(2)  # pylint: disable=no-member
        .get_level3(3)
        .get_level4(4)
        .topic
        == "OpenZWave/2/3/level4/4"
    )
    assert (
        level1.get_level2(2)  # pylint: disable=no-member
        .get_level3(3)
        .get_statistics()
        .topic
        == "OpenZWave/2/3/statistics"
    )


def test_automatic_collections(level1):
    """Test automatic collections."""
    level1.process_message(deque(), {"info": 1})
    level1.process_message(deque(["2"]), {"info": 1})
    level1.process_message(deque(["2", "3"]), {"hello": 1})

    # Test overridden using PLURAL_NAME
    assert list(level1.level_twos()) == [level1.get_level2(2)]

    # Test default name
    assert list(level1.get_level2(2).level3s()) == [level1.get_level2(2).get_level3(3)]


def test_warn_unhandled(level1, caplog):
    """Test warn unhandled."""
    level1.process_message(deque(), {"info": 1})
    level1.process_message(deque(["2"]), {"info": 1})
    level1.process_message(deque(["2", "something"]), {"info": 1})

    assert (
        "Level2 cannot process message OpenZWave/2/something: {'info': 1}"
        in caplog.text
    )


def test_discarder(level1, options, caplog):
    """Test we can discard messages."""
    with patch.object(
        Level1, "create_collections", lambda _: {"command": base.DiscardMessages()}
    ):
        level1 = Level1(options, None, None, None)
        level1.process_message(deque(), {"info": 1})
        level1.process_message(deque(["command"]), {"info": 1})

    assert "cannot process message" not in caplog.text


def test_events(level1, options):
    """Test we can fire events messages."""
    events = []
    options.notify = lambda event, data: events.append((event, data))

    with patch.object(
        Level1,
        "create_collections",
        lambda _: {
            "event": base.EventMessages(
                options, "super_event", lambda topic, msg: topic[0]
            )
        },
    ):
        level1 = Level1(options, None, None, None)
        level1.process_message(deque(), {"info": 1})
        level1.process_message(
            deque(["event", "test-event-type"]), {"data": "for-event"}
        )

    assert len(events) == 1
    event, data = events[0]
    assert event == "super_event"
    assert data == {"event": "test-event-type", "data": {"data": "for-event"}}


def test_repr():
    """Test repr function."""

    class TestNode(base.ZWaveBase):
        """Represent a test node."""

        EVENT_CHANGED = "bla"

    inst = TestNode(None, None, "mock-topic-part", "mock-id")
    assert str(inst) == "<TestNode mock-id>"



"""Setup module for python-openzwave-mqtt."""
from pathlib import Path

from setuptools import find_packages, setup

PROJECT_DIR = Path(__file__).parent.resolve()
README_FILE = PROJECT_DIR / "README.md"
VERSION = "1.0.3"


setup(
    name="python-openzwave-mqtt",
    version=VERSION,
    url="https://github.com/cgarwood/python-openzwave-mqtt",
    download_url="https://github.com/cgarwood/python-openzwave-mqtt",
    author="Charles Garwood",
    author_email="cgarwood@gmail.com",
    description="Converts MQTT messages from qt-openzwave into Python objects and events",
    long_description=README_FILE.read_text(encoding="utf-8"),
    long_description_content_type="text/markdown",
    packages=find_packages(exclude=["test.*", "test"]),
    python_requires=">=3.7",
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Home Automation",
    ],
)



"""Model for Node Statistics."""
from typing import List

from ..base import ZWaveBase
from ..const import EVENT_NODE_ADDED, EVENT_NODE_REMOVED, EVENT_NODE_STATISTICS_CHANGED


class OZWNodeStatistics(ZWaveBase):
    """Model for Node Statistics."""

    EVENT_ADDED = EVENT_NODE_ADDED
    EVENT_CHANGED = EVENT_NODE_STATISTICS_CHANGED
    EVENT_REMOVED = EVENT_NODE_REMOVED

    @property
    def send_count(self) -> int:
        """Return sendCount."""
        return self.data.get("sendCount")

    @property
    def sent_failed(self) -> int:
        """Return sentFailed."""
        return self.data.get("sentFailed")

    @property
    def retries(self) -> int:
        """Return retries."""
        return self.data.get("retries")

    @property
    def received_packets(self) -> int:
        """Return receivedPackets."""
        return self.data.get("receivedPackets")

    @property
    def received_dup_packets(self) -> int:
        """Return receivedDupPackets."""
        return self.data.get("receivedDupPackets")

    @property
    def received_unsolicited(self) -> int:
        """Return receivedUnsolicited."""
        return self.data.get("receivedUnsolicited")

    @property
    def last_sent_time_stamp(self) -> int:
        """Return lastSentTimeStamp."""
        return self.data.get("lastSentTimeStamp")

    @property
    def last_received_time_stamp(self) -> int:
        """Return lastReceivedTimeStamp."""
        return self.data.get("lastReceivedTimeStamp")

    @property
    def last_request_rtt(self) -> int:
        """Return lastRequestRTT."""
        return self.data.get("lastRequestRTT")

    @property
    def average_request_rtt(self) -> int:
        """Return averageRequestRTT."""
        return self.data.get("averageRequestRTT")

    @property
    def last_response_rtt(self) -> int:
        """Return lastResponseRTT."""
        return self.data.get("lastResponseRTT")

    @property
    def average_response_rtt(self) -> int:
        """Return averageResponseRTT."""
        return self.data.get("averageResponseRTT")

    @property
    def quality(self) -> int:
        """Return quality."""
        return self.data.get("quality")

    @property
    def extended_tx_supported(self) -> bool:
        """Return extendedTXSupported."""
        return self.data.get("extendedTXSupported")

    @property
    def tx_time(self) -> int:
        """Return txTime."""
        return self.data.get("txTime")

    @property
    def hops(self) -> int:
        """Return hops."""
        return self.data.get("hops")

    @property
    def rssi1(self) -> str:
        """Return rssi_1."""
        return self.data.get("rssi_1")

    @property
    def rssi2(self) -> str:
        """Return rssi_2."""
        return self.data.get("rssi_2")

    @property
    def rssi3(self) -> str:
        """Return rssi_3."""
        return self.data.get("rssi_3")

    @property
    def rssi4(self) -> str:
        """Return rssi_4."""
        return self.data.get("rssi_4")

    @property
    def rssi5(self) -> str:
        """Return rssi_5."""
        return self.data.get("rssi_5")

    @property
    def route1(self) -> int:
        """Return route_1."""
        return self.data.get("route_1")

    @property
    def route2(self) -> int:
        """Return route_2."""
        return self.data.get("route_2")

    @property
    def route3(self) -> int:
        """Return route_3."""
        return self.data.get("route_3")

    @property
    def route4(self) -> int:
        """Return route_4."""
        return self.data.get("route_4")

    @property
    def ack_channel(self) -> int:
        """Return ackChannel."""
        return self.data.get("ackChannel")

    @property
    def last_tx_channel(self) -> int:
        """Return lastTXChannel."""
        return self.data.get("lastTXChannel")

    @property
    def route_scheme(self) -> str:
        """Return routeScheme."""
        return self.data.get("routeScheme")

    @property
    def route_used(self) -> str:
        """Return routeUsed."""
        return self.data.get("routeUsed")

    @property
    def route_speed(self) -> str:
        """Return routeSpeed."""
        return self.data.get("routeSpeed")

    @property
    def route_tries(self) -> int:
        """Return routeTries."""
        return self.data.get("routeTries")

    @property
    def last_failed_link_from(self) -> int:
        """Return lastFailedLinkFrom."""
        return self.data.get("lastFailedLinkFrom")

    @property
    def last_failed_link_to(self) -> int:
        """Return lastFailedLinkTo."""
        return self.data.get("lastFailedLinkTo")

    @property
    def routes(self) -> List[int]:
        """Return routes."""
        return [
            self.data.get("route_1"),
            self.data.get("route_2"),
            self.data.get("route_3"),
            self.data.get("route_4"),
        ]

    @property
    def rssi(self) -> List[int]:
        """Return rssi."""
        return [
            self.data.get("rssi_1"),
            self.data.get("rssi_2"),
            self.data.get("rssi_3"),
            self.data.get("rssi_4"),
            self.data.get("rssi_5"),
        ]



"""All constants and common types."""
import logging
from enum import Enum, IntEnum

LOGGER = logging.getLogger("openzwavemqtt")

# OZW Events
EVENT_PLACEHOLDER = "missing"
EVENT_COMMAND_CLASS_ADDED = "command_class_added"
EVENT_COMMAND_CLASS_CHANGED = "command_class_changed"
EVENT_COMMAND_CLASS_REMOVED = "command_class_removed"
EVENT_INSTANCE_ADDED = "instance_added"
EVENT_INSTANCE_CHANGED = "instance_changed"
EVENT_INSTANCE_REMOVED = "instance_removed"
EVENT_INSTANCE_EVENT = "instance_event"
EVENT_INSTANCE_STATISTICS_CHANGED = "instance_statistics_changed"
EVENT_INSTANCE_STATUS_CHANGED = "instance_status_changed"
EVENT_NODE_ADDED = "node_added"
EVENT_NODE_CHANGED = "node_changed"
EVENT_NODE_REMOVED = "node_removed"
EVENT_NODE_INSTANCE_ADDED = "node_instance_added"
EVENT_NODE_INSTANCE_CHANGED = "node_instance_changed"
EVENT_NODE_INSTANCE_REMOVED = "node_instance_removed"
EVENT_NODE_ASSOCIATION_ADDED = "node_association_added"
EVENT_NODE_ASSOCIATION_CHANGED = "node_association_changed"
EVENT_NODE_ASSOCIATION_REMOVED = "node_association_removed"
EVENT_NODE_STATISTICS_CHANGED = "node_statistics_changed"
EVENT_VALUE_ADDED = "value_added"
EVENT_VALUE_CHANGED = "value_changed"
EVENT_VALUE_REMOVED = "value_removed"

# Default/empty payload on MQTT messages
EMPTY_PAYLOAD: dict = {}


class OpenZWaveStatus(Enum):
    """Enum with all Status strings for the OZW Daemon."""

    # https://github.com/OpenZWave/qt-openzwave/blob/master/docs/MQTT.md#mqtt-client-status

    OFFLINE = "Offline"
    STARTING = "starting"
    STARTED = "started"
    STOPPED = "stopped"
    READY = "Ready"
    DRIVER_READY = "driverReady"
    DRIVER_FAILED = "driverFailed"
    DRIVER_RESET = "driverReset"
    DRIVER_REMOVED = "driverRemoved"
    DRIVER_ALL_NODES_QUERIED_SOME_DEAD = "driverAllNodesQueriedSomeDead"
    DRIVER_ALL_NODES_QUERIED = "driverAllNodesQueried"
    DRIVER_AWAKE_NODES_QUERIED_SOME_DEAD = "driverAwakeNodesQueried"
    UNKNOWN = None


# Tuple with all states indicating OZW is ready
OZW_READY_STATES = (
    OpenZWaveStatus.DRIVER_AWAKE_NODES_QUERIED_SOME_DEAD,
    OpenZWaveStatus.DRIVER_ALL_NODES_QUERIED_SOME_DEAD,
    OpenZWaveStatus.DRIVER_ALL_NODES_QUERIED,
)


class ValueGenre(Enum):
    """Enum with all diferent Value genres."""

    USER = "User"
    SYSTEM = "System"
    BASIC = "Basic"
    UNKNOWN = None


class ValueType(Enum):
    """Enum with all diferent Value types."""

    BOOL = "Bool"
    LIST = "List"
    STRING = "String"
    DECIMAL = "Decimal"
    BYTE = "Byte"
    INT = "Int"
    BUTTON = "Button"
    UNKNOWN = None


class ValueIndex(IntEnum):
    """Enum with all (known/used) Value indexes."""

    # Alarm
    ALARM_TYPE = 0
    ALARM_LEVEL = 1
    ALARM_ACCESS_CONTROL = 9
    # BarrierOperator
    BARRIER_OPERATOR_LABEL = 1
    # DoorLock
    DOOR_LOCK_LOCK = 0
    # Meter
    METER_POWER = 2
    METER_RESET = 257
    # SensorMultilevel
    SENSOR_MULTILEVEL_AIR_TEMPERATURE = 1
    SENSOR_MULTILEVEL_GENERAL_PURPOSE = 2
    SENSOR_MULTILEVEL_ILLUMINANCE = 3
    SENSOR_MULTILEVEL_POWER = 4
    SENSOR_MULTILEVEL_HUMIDITY = 5
    SENSOR_MULTILEVEL_VELOCITY = 6
    SENSOR_MULTILEVEL_DIRECTION = 7
    SENSOR_MULTILEVEL_ATMOSPHERIC_PRESSURE = 8
    SENSOR_MULTILEVEL_BAROMETIC_PRESSURE = 9
    SENSOR_MULTILEVEL_SOLAR_RADIATION = 10
    SENSOR_MULTILEVEL_DEW_POINT = 11
    SENSOR_MULTILEVEL_RAIN_RATE = 12
    SENSOR_MULTILEVEL_TIDE_LEVEL = 13
    SENSOR_MULTILEVEL_WEIGHT = 14
    SENSOR_MULTILEVEL_VOLTAGE = 15
    SENSOR_MULTILEVEL_CURRENT = 16
    SENSOR_MULTILEVEL_CARBON_DIOXIDE = 17
    SENSOR_MULTILEVEL_AIR_FLOW = 18
    SENSOR_MULTILEVEL_TANK_CAPACITY = 19
    SENSOR_MULTILEVEL_DISTANCE = 20
    SENSOR_MULTILEVEL_ANGLE_POSITION = 21
    SENSOR_MULTILEVEL_ROTATION = 22
    SENSOR_MULTILEVEL_WATER_TEMPERATURE = 23
    SENSOR_MULTILEVEL_SOIL_TEMPERATURE = 24
    SENSOR_MULTILEVEL_SEISMIC_INTENSITY = 25
    SENSOR_MULTILEVEL_SEISMIC_MAGNITUDE = 26
    SENSOR_MULTILEVEL_ULTRAVIOLET = 27
    SENSOR_MULTILEVEL_ELECTRICAL_RESISTIVITY = 28
    SENSOR_MULTILEVEL_ELECTRICAL_CONDUCTIVITY = 29
    SENSOR_MULTILEVEL_LOUDNESS = 30
    SENSOR_MULTILEVEL_MOISTURE = 31
    SENSOR_MULTILEVEL_FREQUENCY = 32
    SENSOR_MULTILEVEL_TIME = 33
    SENSOR_MULTILEVEL_TARGET_TEMPERATURE = 34
    SENSOR_MULTILEVEL_PARTICULATE_MATTER_25 = 35
    SENSOR_MULTILEVEL_FORMALDEHYDE_CH20_LEVEL = 36
    SENSOR_MULTILEVEL_RADON_CONCENTRATION = 37
    SENSOR_MULTILEVEL_METHANE_DENSITY = 38
    SENSOR_MULTILEVEL_VOLATILE_ORGANIC_COMPOUND = 39
    SENSOR_MULTILEVEL_CARBON_MONOXIDE = 40
    SENSOR_MULTILEVEL_SOIL_HUMIDITY = 41
    SENSOR_MULTILEVEL_SOIL_REACTIVITY = 42
    SENSOR_MULTILEVEL_SOIL_SALINITY = 43
    SENSOR_MULTILEVEL_HEART_RATE = 44
    SENSOR_MULTILEVEL_BLOOD_PRESSURE = 45
    SENSOR_MULTILEVEL_MUSCLE_MASS = 46
    SENSOR_MULTILEVEL_FAT_MASS = 47
    SENSOR_MULTILEVEL_BONE_MASS = 48
    SENSOR_MULTILEVEL_TOTAL_BODY_WATER = 49
    SENSOR_MULTILEVEL_BASIC_METABOLIC_RATE = 50
    SENSOR_MULTILEVEL_BODY_MASS_INDEX = 51
    SENSOR_MULTILEVEL_X_AXIS_ACCELERATION = 52
    SENSOR_MULTILEVEL_Y_AXIS_ACCELERATION = 53
    SENSOR_MULTILEVEL_Z_AXIS_ACCELERATION = 54
    SENSOR_MULTILEVEL_SMOKE_DENSITY = 55
    SENSOR_MULTILEVEL_WATER_FLOW = 56
    SENSOR_MULTILEVEL_WATER_PRESSURE = 57
    SENSOR_MULTILEVEL_RF_SIGNAL_STRENGTH = 58
    SENSOR_MULTILEVEL_PARTICULATE_MATTER_10 = 59
    SENSOR_MULTILEVEL_RESPIRATORY_RATE = 60
    SENSOR_MULTILEVEL_RELATIVE_MODULATION = 61
    SENSOR_MULTILEVEL_BOILER_WATER_TEMPERATURE = 62
    SENSOR_MULTILEVEL_DOMESTIC_HOT_WATER_TEMPERATURE = 63
    SENSOR_MULTILEVEL_OUTSIDE_TEMPERATURE = 64
    SENSOR_MULTILEVEL_EXHAUST_TEMPERATURE = 65
    SENSOR_MULTILEVEL_WATER_CHLORINE = 66
    SENSOR_MULTILEVEL_WATER_ACIDITY = 67
    SENSOR_MULTILEVEL_WATER_OXIDATION_REDUCTION_POTENTIAL = 68
    SENSOR_MULTILEVEL_HEART_RATE_LF_HF_RATIO = 69
    SENSOR_MULTILEVEL_MOTION_DIRECTION = 70
    SENSOR_MULTILEVEL_APPLIED_FORCE = 71
    SENSOR_MULTILEVEL_RETURN_AIR_TEMPERATURE = 72
    SENSOR_MULTILEVEL_SUPPLY_AIR_TEMPERATURE = 73
    SENSOR_MULTILEVEL_CONDENSER_COIL_TEMPERATURE = 74
    SENSOR_MULTILEVEL_EVAPORATOR_COIL_TEMPERATURE = 75
    SENSOR_MULTILEVEL_LIQUID_LINE_TEMPERATURE = 76
    SENSOR_MULTILEVEL_DISCHARGE_LINE_TEMPERATURE = 77
    SENSOR_MULTILEVEL_SUCTION = 78
    SENSOR_MULTILEVEL_DISCHARGE = 79
    SENSOR_MULTILEVEL_DEFROST_TEMPERATURE = 80
    SENSOR_MULTILEVEL_OZONE = 81
    SENSOR_MULTILEVEL_SULFUR_DIOXIDE = 82
    SENSOR_MULTILEVEL_NITROGEN_DIOXIDE = 83
    SENSOR_MULTILEVEL_AMMONIA = 84
    SENSOR_MULTILEVEL_LEAD = 85
    SENSOR_MULTILEVEL_PARTICULATE_MATTER_1 = 86
    # Color
    SWITCH_COLOR_COLOR = 0
    SWITCH_COLOR_CHANNELS = 2
    # SwitchMultilevel
    SWITCH_MULTILEVEL_LEVEL = 0
    SWITCH_MULTILEVEL_BRIGHT = 1
    SWITCH_MULTILEVEL_DIM = 2
    SWITCH_MULTILEVEL_DURATION = 5
    # Notification
    NOTIFICATION_SMOKE_ALARM = 1
    NOTIFICATION_CARBON_MONOOXIDE = 2
    NOTIFICATION_CARBON_DIOXIDE = 3
    NOTIFICATION_HEAT = 4
    NOTIFICATION_WATER = 5
    NOTIFICATION_ACCESS_CONTROL = 6
    NOTIFICATION_HOME_SECURITY = 7
    NOTIFICATION_POWER_MANAGEMENT = 8
    NOTIFICATION_SYSTEM = 9
    NOTIFICATION_EMERGENCY = 10
    NOTIFICATION_CLOCK = 11
    NOTIFICATION_APPLIANCE = 12
    NOTIFICATION_HOME_HEALTH = 13
    NOTIFICATION_SIREN = 14
    NOTIFICATION_WATER_VALVE = 15
    NOTIFICATION_WEATHER = 16
    NOTIFICATION_IRRIGATION = 17
    NOTIFICATION_GAS = 18
    UNKNOWN = 999


class CommandClass(IntEnum):
    """Enum with all known CommandClasses."""

    ALARM = 113
    SENSOR_ALARM = 156
    SILENCE_ALARM = 157
    SWITCH_ALL = 39
    ANTITHEFT = 93
    ANTITHEFT_UNLOCK = 126
    APPLICATION_CAPABILITY = 87
    APPLICATION_STATUS = 34
    ASSOCIATION = 133
    ASSOCIATION_COMMAND_CONFIGURATION = 155
    ASSOCIATION_GRP_INFO = 89
    AUTHENTICATION = 161
    AUTHENTICATION_MEDIA_WRITE = 162
    BARRIER_OPERATOR = 102
    BASIC = 32
    BASIC_TARIFF_INFO = 54
    BASIC_WINDOW_COVERING = 80
    BATTERY = 128
    SENSOR_BINARY = 48
    SWITCH_BINARY = 37
    SWITCH_TOGGLE_BINARY = 40
    CLIMATE_CONTROL_SCHEDULE = 70
    CENTRAL_SCENE = 91
    CLOCK = 129
    SWITCH_COLOR = 51
    CONFIGURATION = 112
    CONTROLLER_REPLICATION = 33
    CRC_16_ENCAP = 86
    DCP_CONFIG = 58
    DCP_MONITOR = 59
    DEVICE_RESET_LOCALLY = 90
    DOOR_LOCK = 98
    DOOR_LOCK_LOGGING = 76
    ENERGY_PRODUCTION = 144
    ENTRY_CONTROL = 111
    FIRMWARE_UPDATE_MD = 122
    GENERIC_SCHEDULE = 163
    GEOGRAPHIC_LOCATION = 140
    GROUPING_NAME = 123
    HAIL = 130
    HRV_STATUS = 55
    HRV_CONTROL = 57
    HUMIDITY_CONTROL_MODE = 109
    HUMIDITY_CONTROL_OPERATING_STATE = 110
    HUMIDITY_CONTROL_SETPOINT = 100
    INCLUSION_CONTROLLER = 116
    INDICATOR = 135
    IP_ASSOCIATION = 92
    IP_CONFIGURATION = 154
    IR_REPEATER = 160
    IRRIGATION = 107
    LANGUAGE = 137
    LOCK = 118
    MAILBOX = 105
    MANUFACTURER_PROPRIETARY = 145
    MANUFACTURER_SPECIFIC = 114
    MARK = 239
    METER = 50
    METER_TBL_CONFIG = 60
    METER_TBL_MONITOR = 61
    METER_TBL_PUSH = 62
    MTP_WINDOW_COVERING = 81
    MULTI_CHANNEL = 96
    MULTI_CHANNEL_ASSOCIATION = 142
    MULTI_CMD = 143
    SENSOR_MULTILEVEL = 49
    SWITCH_MULTILEVEL = 38
    SWITCH_TOGGLE_MULTILEVEL = 41
    NETWORK_MANAGEMENT_BASIC = 77
    NETWORK_MANAGEMENT_INCLUSION = 52
    NETWORK_MANAGEMENT_INSTALLATION_MAINTENANCE = 103
    NETWORK_MANAGEMENT_PRIMARY = 84
    NETWORK_MANAGEMENT_PROXY = 82
    NO_OPERATION = 0
    NODE_NAMING = 119
    NODE_PROVISIONING = 120
    NOTIFICATION = 113
    POWERLEVEL = 115
    PREPAYMENT = 63
    PREPAYMENT_ENCAPSULATION = 65
    PROPRIETARY = 136
    PROTECTION = 117
    METER_PULSE = 53
    RATE_TBL_CONFIG = 72
    RATE_TBL_MONITOR = 73
    REMOTE_ASSOCIATION_ACTIVATE = 124
    REMOTE_ASSOCIATION = 125
    SCENE_ACTIVATION = 43
    SCENE_ACTUATOR_CONF = 44
    SCENE_CONTROLLER_CONF = 45
    SCHEDULE = 83
    SCHEDULE_ENTRY_LOCK = 78
    SCREEN_ATTRIBUTES = 147
    SCREEN_MD = 146
    SECURITY = 152
    SECURITY_2 = 159
    SECURITY_SCHEME0_MARK = 61696
    SENSOR_CONFIGURATION = 158
    SIMPLE_AV_CONTROL = 148
    SOUND_SWITCH = 121
    SUPERVISION = 108
    TARIFF_CONFIG = 74
    TARIFF_TBL_MONITOR = 75
    THERMOSTAT_FAN_MODE = 68
    THERMOSTAT_FAN_STATE = 69
    THERMOSTAT_MODE = 64
    THERMOSTAT_OPERATING_STATE = 66
    THERMOSTAT_SETBACK = 71
    THERMOSTAT_SETPOINT = 67
    TIME = 138
    TIME_PARAMETERS = 139
    TRANSPORT_SERVICE = 85
    USER_CODE = 99
    VERSION = 134
    WAKE_UP = 132
    WINDOW_COVERING = 106
    ZIP = 35
    ZIP_6LOWPAN = 79
    ZIP_GATEWAY = 95
    ZIP_NAMING = 104
    ZIP_ND = 88
    ZIP_PORTAL = 97
    ZWAVEPLUS_INFO = 94
    UNKNOWN = 0


# DeviceTypes Generic/Specific

SPECIFIC_TYPE_NOT_USED = 0  # Available in all Generic types
GENERIC_TYPE_AV_CONTROL_POINT = 3
SPECIFIC_TYPE_DOORBELL = 18
SPECIFIC_TYPE_SATELLITE_RECEIVER = 4
SPECIFIC_TYPE_SATELLITE_RECEIVER_V2 = 17

GENERIC_TYPE_DISPLAY = 4
SPECIFIC_TYPE_SIMPLE_DISPLAY = 1

GENERIC_TYPE_ENTRY_CONTROL = 64
SPECIFIC_TYPE_DOOR_LOCK = 1
SPECIFIC_TYPE_ADVANCED_DOOR_LOCK = 2
SPECIFIC_TYPE_SECURE_KEYPAD_DOOR_LOCK = 3
SPECIFIC_TYPE_SECURE_KEYPAD_DOOR_LOCK_DEADBOLT = 4
SPECIFIC_TYPE_SECURE_DOOR = 5
SPECIFIC_TYPE_SECURE_GATE = 6
SPECIFIC_TYPE_SECURE_BARRIER_ADDON = 7
SPECIFIC_TYPE_SECURE_BARRIER_OPEN_ONLY = 8
SPECIFIC_TYPE_SECURE_BARRIER_CLOSE_ONLY = 9
SPECIFIC_TYPE_SECURE_LOCKBOX = 10
SPECIFIC_TYPE_SECURE_KEYPAD = 11

GENERIC_TYPE_GENERIC_CONTROLLER = 1
SPECIFIC_TYPE_PORTABLE_CONTROLLER = 1
SPECIFIC_TYPE_PORTABLE_SCENE_CONTROLLER = 2
SPECIFIC_TYPE_PORTABLE_INSTALLER_TOOL = 3
SPECIFIC_TYPE_REMOTE_CONTROL_AV = 4
SPECIFIC_TYPE_REMOTE_CONTROL_SIMPLE = 6

GENERIC_TYPE_METER = 49
SPECIFIC_TYPE_SIMPLE_METER = 1
SPECIFIC_TYPE_ADV_ENERGY_CONTROL = 2
SPECIFIC_TYPE_WHOLE_HOME_METER_SIMPLE = 3

GENERIC_TYPE_METER_PULSE = 48

GENERIC_TYPE_NON_INTEROPERABLE = 255

GENERIC_TYPE_REPEATER_SLAVE = 15
SPECIFIC_TYPE_REPEATER_SLAVE = 1
SPECIFIC_TYPE_VIRTUAL_NODE = 2

GENERIC_TYPE_SECURITY_PANEL = 23
SPECIFIC_TYPE_ZONED_SECURITY_PANEL = 1

GENERIC_TYPE_SEMI_INTEROPERABLE = 80
SPECIFIC_TYPE_ENERGY_PRODUCTION = 1

GENERIC_TYPE_SENSOR_ALARM = 161
SPECIFIC_TYPE_ADV_ZENSOR_NET_ALARM_SENSOR = 5
SPECIFIC_TYPE_ADV_ZENSOR_NET_SMOKE_SENSOR = 10
SPECIFIC_TYPE_BASIC_ROUTING_ALARM_SENSOR = 1
SPECIFIC_TYPE_BASIC_ROUTING_SMOKE_SENSOR = 6
SPECIFIC_TYPE_BASIC_ZENSOR_NET_ALARM_SENSOR = 3
SPECIFIC_TYPE_BASIC_ZENSOR_NET_SMOKE_SENSOR = 8
SPECIFIC_TYPE_ROUTING_ALARM_SENSOR = 2
SPECIFIC_TYPE_ROUTING_SMOKE_SENSOR = 7
SPECIFIC_TYPE_ZENSOR_NET_ALARM_SENSOR = 4
SPECIFIC_TYPE_ZENSOR_NET_SMOKE_SENSOR = 9
SPECIFIC_TYPE_ALARM_SENSOR = 11

GENERIC_TYPE_SENSOR_BINARY = 32
SPECIFIC_TYPE_ROUTING_SENSOR_BINARY = 1

GENERIC_TYPE_SENSOR_MULTILEVEL = 33
SPECIFIC_TYPE_ROUTING_SENSOR_MULTILEVEL = 1
SPECIFIC_TYPE_CHIMNEY_FAN = 2

GENERIC_TYPE_STATIC_CONTROLLER = 2
SPECIFIC_TYPE_PC_CONTROLLER = 1
SPECIFIC_TYPE_SCENE_CONTROLLER = 2
SPECIFIC_TYPE_STATIC_INSTALLER_TOOL = 3
SPECIFIC_TYPE_SET_TOP_BOX = 4
SPECIFIC_TYPE_SUB_SYSTEM_CONTROLLER = 5
SPECIFIC_TYPE_TV = 6
SPECIFIC_TYPE_GATEWAY = 7

GENERIC_TYPE_SWITCH_BINARY = 16
SPECIFIC_TYPE_POWER_SWITCH_BINARY = 1
SPECIFIC_TYPE_SCENE_SWITCH_BINARY = 3
SPECIFIC_TYPE_POWER_STRIP = 4
SPECIFIC_TYPE_SIREN = 5
SPECIFIC_TYPE_VALVE_OPEN_CLOSE = 6
SPECIFIC_TYPE_COLOR_TUNABLE_BINARY = 2
SPECIFIC_TYPE_IRRIGATION_CONTROLLER = 7

GENERIC_TYPE_SWITCH_MULTILEVEL = 17
SPECIFIC_TYPE_CLASS_A_MOTOR_CONTROL = 5
SPECIFIC_TYPE_CLASS_B_MOTOR_CONTROL = 6
SPECIFIC_TYPE_CLASS_C_MOTOR_CONTROL = 7
SPECIFIC_TYPE_MOTOR_MULTIPOSITION = 3
SPECIFIC_TYPE_POWER_SWITCH_MULTILEVEL = 1
SPECIFIC_TYPE_SCENE_SWITCH_MULTILEVEL = 4
SPECIFIC_TYPE_FAN_SWITCH = 8
SPECIFIC_TYPE_COLOR_TUNABLE_MULTILEVEL = 2

GENERIC_TYPE_SWITCH_REMOTE = 18
SPECIFIC_TYPE_REMOTE_BINARY = 1
SPECIFIC_TYPE_REMOTE_MULTILEVEL = 2
SPECIFIC_TYPE_REMOTE_TOGGLE_BINARY = 3
SPECIFIC_TYPE_REMOTE_TOGGLE_MULTILEVEL = 4

GENERIC_TYPE_SWITCH_TOGGLE = 19
SPECIFIC_TYPE_SWITCH_TOGGLE_BINARY = 1
SPECIFIC_TYPE_SWITCH_TOGGLE_MULTILEVEL = 2

GENERIC_TYPE_THERMOSTAT = 8
SPECIFIC_TYPE_SETBACK_SCHEDULE_THERMOSTAT = 3
SPECIFIC_TYPE_SETBACK_THERMOSTAT = 5
SPECIFIC_TYPE_SETPOINT_THERMOSTAT = 4
SPECIFIC_TYPE_THERMOSTAT_GENERAL = 2
SPECIFIC_TYPE_THERMOSTAT_GENERAL_V2 = 6
SPECIFIC_TYPE_THERMOSTAT_HEATING = 1

GENERIC_TYPE_VENTILATION = 22
SPECIFIC_TYPE_RESIDENTIAL_HRV = 1

GENERIC_TYPE_WINDOWS_COVERING = 9
SPECIFIC_TYPE_SIMPLE_WINDOW_COVERING = 1

GENERIC_TYPE_ZIP_NODE = 21
SPECIFIC_TYPE_ZIP_ADV_NODE = 2
SPECIFIC_TYPE_ZIP_TUN_NODE = 1

GENERIC_TYPE_WALL_CONTROLLER = 24
SPECIFIC_TYPE_BASIC_WALL_CONTROLLER = 1

GENERIC_TYPE_NETWORK_EXTENDER = 5
SPECIFIC_TYPE_SECURE_EXTENDER = 1

GENERIC_TYPE_APPLIANCE = 6
SPECIFIC_TYPE_GENERAL_APPLIANCE = 1
SPECIFIC_TYPE_KITCHEN_APPLIANCE = 2
SPECIFIC_TYPE_LAUNDRY_APPLIANCE = 3

GENERIC_TYPE_SENSOR_NOTIFICATION = 7
SPECIFIC_TYPE_NOTIFICATION_SENSOR = 1
